#!/usr/bin/python3

## Proof of Space-Time prover

import math, common, ast, sys

# For the ratio between storing and power:
# Note that for a production ready system hashing speedup (as designed for Bitcoin in [AsicBoost - A Speedup for Bitcoin Mining, Dr. Timo Hanke, March 31, 2016 (rev. 5)](https://arxiv.org/pdf/1604.00575.pdf)) should be considered.

## Initialization phase: 1. Receive a random bitstring from the verifier.

protocolInitializationPhaseId = input('protocolInitializationPhaseId: ')

## Initialization phase: 2. Generate in a table the data that the prover has to store.
# To ensure that this data is cheaper to store than to generate and to allow public verifiability by requiring the stored data to be Proof of Works.
# For the ease of this implementation, the data are stored in the memory instead of the disk.

STORED_NONCES = []
counter = 0
lastCounter = 0

# We enforce `STORED_NONCES` to have its length being a power of 2, to ease Merkle tree implementation.
while len(STORED_NONCES) < common.ENTRIES_NUMBER:
    hashed = common.hash(protocolInitializationPhaseId + str(counter))
    if common.doesHashMatchDifficulty(hashed):
        deltaCounter = counter - lastCounter
        STORED_NONCES += [deltaCounter]
        lastCounter = counter
        print(len(STORED_NONCES), common.ENTRIES_NUMBER, file=sys.stderr)
    counter += 1

while True:
    print()
    ## Execution phase: 1. Receive a random bitstring from the verifier.

    protocolExecutionPhaseId = input('protocolExecutionPhaseId: ')

    ## Execution phase: 2. The prover commits to the table contents given this random challenge.

    def pairwise(iterable):
        "s -> (s0, s1), (s2, s3), (s4, s5), ..."
        a = iter(iterable)
        return zip(a, a)

    def getNextMerkleTreeLevel(nodes):
        return [common.hash(leftNode + rightNode) for leftNode, rightNode in pairwise(nodes)]

    merkleTreeLevel = []
    counter = 0
    for deltaCounter in STORED_NONCES:
        counter += deltaCounter
        hashed = common.hash(protocolInitializationPhaseId + protocolExecutionPhaseId + str(counter))
        merkleTreeLevel += [hashed]

    merkleTreeLevels = [merkleTreeLevel]

    while len(merkleTreeLevels[-1]) > 1:
        nextMerkleTreeLevel = getNextMerkleTreeLevel(merkleTreeLevels[-1])
        merkleTreeLevels += [nextMerkleTreeLevel]

    merkleTreeRoot = merkleTreeLevels[-1][0]

    print(f'{merkleTreeRoot=}')

    ## Execution phase: 3. Receive a random set of indexes from the verifier.

    with open(input('indexesRequest: ')) as f:
        indexesRequest = ast.literal_eval(f.read())

    ## Execution phase: 4. Respond with the corresponding table entries and commitment openings.

    # First computing indexes of fields required in the different levels.
    # Then just transmit the hashes of these indexes, as the verifier can compute these indexes on its own.
    # Then proceed to the Merkle tree root validation. Designing this first to make sure transmit in the easiest way make sense.

    # For verifying treat level by level

    # Just sending a list of hashes that the verifier fills its data structure with, as the prover generated this list.
    # So have to just verify that there's no such attack when filling the first level.
    # Verify especially my parsing even with the use of `ast.literal_eval` like the type and the size of entries and possibly subentries.

    # Could also proceed index by index by providing entry and siblings up to the root for each index.
    # Could use page 28 of FlyClient: Super-Light Clients for Cryptocurrencies for the implementation.

    verificationMerkleTreeLevels = [{} for _ in range(common.LEVELS)]

    for indexRequest in indexesRequest:
        otherIndex = common.otherIndex(indexRequest)
        for index in [indexRequest, otherIndex]:
            verificationMerkleTreeLevels[0][index] = merkleTreeLevels[0][index]

    # Should rename `entries`, as they aren't *entries* for higher levels.
    entries = []
    for verificationMerkleTreeLevelsIndex, verificationMerkleTreeLevel in enumerate(verificationMerkleTreeLevels[:-1]):
        for index in verificationMerkleTreeLevel:
            otherIndex = common.otherIndex(index)
            leftIndex, rightIndex = sorted([index, otherIndex])
            leftHash, rightHash = [merkleTreeLevels[verificationMerkleTreeLevelsIndex][i] for i in [leftIndex, rightIndex]]
            otherHash = merkleTreeLevels[verificationMerkleTreeLevelsIndex][otherIndex]
            verificationMerkleTreeLevels[verificationMerkleTreeLevelsIndex + 1][leftIndex // 2] = common.hash(leftHash + rightHash)
            if verificationMerkleTreeLevelsIndex > 0 or leftIndex == index:
                if verificationMerkleTreeLevelsIndex == 0:
                    leftNonce = sum(STORED_NONCES[:leftIndex + 1])
                    rightNonce = leftNonce + STORED_NONCES[rightIndex]
                    entries += [leftNonce, rightNonce]
                else:
                    entries += [otherHash]

    entriesFilePath = 'entries.txt'

    with open(entriesFilePath, 'w') as f:
        f.write(str(entries))

    print(f'entries: {entriesFilePath}')
